/*********************************************************************

時計型の妖怪音声おもちゃ(白)用 妖怪コードの赤外線リモコン制御ソフト

				IR to Yok Code Converter for Arduino

　　　お手持ちのシャープのテレビ用の赤外線リモコンの数字ボタンで
　　　４ケタの妖怪コードを入力すると動作します。
　　　また、チャンネル選局の上ボタンで、次のコードを呼び出します。

本ソースリストおよびソフトウェアは、ライセンスフリーです。
利用、編集、再配布等が自由に行えますが、著作権表示の改変は禁止します。

                               Copyright (c) 2015 Wataru KUNINO
                               http://www.geocities.jp/bokunimowakaru/
*********************************************************************/
/*
・赤外線リモコン受信モジュールを Arduino Digital 2 に接続します。
・妖怪おもちゃ内部基板のSW3～6を Arduino Digital 3～6 に接続します。
・赤外線リモコンを使用しない場合は#defineのIrRemoENを0にします。

注意点：
　赤外線リモコンを使用する(IrRemoENが1)状態の時は低消費電力モードで動作します。
　この場合、シリアルからのYokCodeを受信することが出来ません。
　シリアルから妖怪ウォッチを制御する場合は、IrRemoENを0に設定してください。
　Arduinoの割込みの制約上、実メダルからパターンを取得することが出来ません。

参考文献：
　妖怪コードに関する情報は下記のサイトを参考にしました。
　http://yokai.kaathemachine.com/
　
　白の範囲
　2382～3511

　零式の範囲
　0: 339～ 351
　1: 857～ 863
　2:1617～1617
　3:1619～1619
　4:1624～1631
　5:1873～1887
　6:2139～2139
　7:2141～2142
　8:3090～3281
　9:3484～3487
　A:3505～3508
　B:3510～3522
　C:3665～3679
　D:3921～3934
*/

#define SW3_PIN			3						// 妖怪おもちゃ内SW3の接続先
#define SW4_PIN			4						// 妖怪おもちゃ内SW4の接続先
#define SW5_PIN			5						// 妖怪おもちゃ内SW5の接続先
#define SW6_PIN			6						// 妖怪おもちゃ内SW6の接続先
#define LED_PIN			13						// 動作確認用LEDのDigitalピン番号
#define IR_VCC_PIN		8						// 赤外線モジュール用電源ピン
//#define IR_PIN		2 (固定)				// 赤外線センサ(Digital 2 固定)

#define IrRemoEN		1						// 赤外線リモコン入力 1:使用
#define SerialEN		1						// シリアル入力 1:使用
#define SymWait			10						// YokCode出力間隔
#define IR_DATA_LEN		8						// リモコンコードのデータ長(byte)
#define AutoPowerSave	8						// 自動で電源を切るまでの時間(8秒倍)

#define YokTypeWhite	1						// 妖怪おもちゃの形式 1:白
#define YokTypeZero		0						// 妖怪おもちゃの形式 1:零式

#include <avr/sleep.h>
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))

volatile int keyValue = 0;						// ユーザ入力用バッファ
volatile byte keyLength = 0;					// ユーザ入力文字数
volatile byte ardIsrSrc = 0xFF;					// 割込み状態
volatile boolean yokCodeMedal = 0;				// YokCodeメダルの状態 0:空 1:挿入
volatile byte time = 0;							// 時間カウント用

int yokCode;									// 初期のYokCode(可変値)
int YokCodeMin,YokCodeMax;						// YokCodeの上下限値(固定値)

void yok_init(){
// YokCodeの初期値と上限値YokCodeMaxと下限値YokCodeMinの設定
	if(YokTypeWhite){
		yokCode = 2913;							// 0xB61 2913=地縛霊猫
		YokCodeMin = 2382;						// 0x94E
		YokCodeMax = 3511;						// 0xDB7
	}
	if(YokTypeZero){
		yokCode = 339;
		YokCodeMin = 339;
		YokCodeMax = 3934;
	}
}

void port_off(){
// 全SWポートをオープンにする
	pinMode(SW3_PIN,INPUT);
	pinMode(SW4_PIN,INPUT);
	pinMode(SW5_PIN,INPUT);
	pinMode(SW6_PIN,INPUT);
}

void port_on(){
// 全SWポートをLOW出力につする
	pinMode(SW3_PIN,OUTPUT);
	pinMode(SW4_PIN,OUTPUT);
	pinMode(SW5_PIN,OUTPUT);
	pinMode(SW6_PIN,OUTPUT);
	digitalWrite(SW3_PIN,LOW);
	digitalWrite(SW4_PIN,LOW);
	digitalWrite(SW5_PIN,LOW);
	digitalWrite(SW6_PIN,LOW);
}

void port_out(byte in){
// ポートにYokCodeを出力する
	digitalWrite(SW3_PIN,in>>0 & 0x01);
	digitalWrite(SW4_PIN,in>>1 & 0x01);
	digitalWrite(SW5_PIN,in>>2 & 0x01);
	digitalWrite(SW6_PIN,in>>3 & 0x01);
}

void code_out(int code,boolean dir){
	byte out,i;
	
	port_on();
	if(SerialEN){
		Serial.print("yokCode = "); Serial.print(code,DEC);
		Serial.print(", dir = ");  Serial.println(dir,DEC);
	}
	if(!dir){
		port_out(0x0A); delay(SymWait);
		port_out(0x00); delay(SymWait);
	}
	for(i=0;i<3;i++){
		if(dir){	// dir=1:メダル挿入
			out = (byte)(code>>(4*(2-i))) & 0x0F;
		}else{		// dir=0:メダル取り出し
			out = (byte)(code>>(4*i)) & 0x0F;
		}
		port_out(out);
		delay(SymWait);
		port_out(0x00); delay(SymWait);
	}
	if(dir){
		port_out(0x0A); delay(SymWait);
		port_out(0x00); delay(SymWait);
	}
	port_off();
}

int code_increment(int code){
	byte data[3],flag=1,i;
	
	if(code >= YokCodeMax) code = YokCodeMin;
	else if(YokTypeZero && code ==  351 ) code =  857;	// 0 -> 1
	else if(YokTypeZero && code ==  863 ) code = 1617;	// 1 -> 2
	else if(YokTypeZero && code == 1617 ) code = 1619;	// 2 -> 3
	else if(YokTypeZero && code == 1619 ) code = 1624;	// 3 -> 4
	else if(YokTypeZero && code == 1631 ) code = 1873;	// 4 -> 5
	else if(YokTypeZero && code == 1887 ) code = 2139;	// 5 -> 6
	else if(YokTypeZero && code == 2139 ) code = 2141;	// 6 -> 7
	else if(YokTypeZero && code == 2142 ) code = 3090;	// 7 -> 8
	else if(YokTypeZero && code == 3281 ) code = 3484;	// 8 -> 9
	else if(YokTypeZero && code == 3487 ) code = 3505;	// 9 -> A
	else if(YokTypeZero && code == 3508 ) code = 3510;	// A -> B
	else if(YokTypeZero && code == 3522 ) code = 3665;	// B -> C
	else if(YokTypeZero && code == 3679 ) code = 3921;	// C -> D
	else{
		for(i=0;i<3;i++) data[i]=(byte)(code>>(4*(int)i)) & 0x0F;
		while(flag){					// 適切なコードが見つかるまで繰り返す
			data[0]++;
			if(data[0] > 0x0F){			// コードの桁上がり
				data[0] = 0x00;
				data[1]++;
				if(data[1] > 0x0F){		// 種類コード(1列目)の桁上がり
					data[1] = 0x00;
					if(YokTypeWhite) data[2]+= 0x02; else data[2]++;
				}
			}
			if(	data[0]!=data[2] &&		// 不一致コードを確認する
				data[0]!=data[1] &&
				data[0]!=0x00 &&
				data[0]!=0x05 &&
				data[0]!=0x0A
			) flag=0;
		}
		code=0;
		for(i=0;i<3;i++){
			code *= 16;
			code += (int)data[2-i];
		}
	}
	if(SerialEN){
		Serial.print("IncCode = 0x");
		for(i=0;i<3;i++) Serial.print(data[2-i],HEX);
		Serial.print(" -> "); Serial.println(code,DEC);
	}
	return(code);
}

int code_decrement(int code){
	byte data[3],flag=1,i;
	
	if(code <= YokCodeMin) return(YokCodeMax);
	else if(YokTypeZero && code ==  857 ) code =  351;	// 1 -> 0
	else if(YokTypeZero && code == 1617 ) code =  863;	// 2 -> 1
	else if(YokTypeZero && code == 1619 ) code = 1617;	// 3 -> 2
	else if(YokTypeZero && code == 1624 ) code = 1619;	// 4 -> 3
	else if(YokTypeZero && code == 1873 ) code = 1631;	// 5 -> 4
	else if(YokTypeZero && code == 2139 ) code = 1887;	// 6 -> 5
	else if(YokTypeZero && code == 2141 ) code = 2139;	// 7 -> 6
	else if(YokTypeZero && code == 3090 ) code = 2142;	// 8 -> 7
	else if(YokTypeZero && code == 3484 ) code = 3281;	// 9 -> 8
	else if(YokTypeZero && code == 3505 ) code = 3487;	// A -> 9
	else if(YokTypeZero && code == 3510 ) code = 3508;	// B -> A
	else if(YokTypeZero && code == 3665 ) code = 3522;	// C -> B
	else if(YokTypeZero && code == 3921 ) code = 3679;	// D -> C
	else{
		for(i=0;i<3;i++) data[i]=(byte)(code>>(4*(int)i)) & 0x0F;
		while(flag){
			data[0]--;
			if(data[0] > 0x0F){
				data[0] = 0x0F;
				data[1]--;
				if(data[1] > 0x0F){
					data[1] = 0x0F;
					if(YokTypeWhite) data[2]-= 0x02; else data[2]--;
				}
			}
			if(	data[0]!=data[2] &&
				data[0]!=data[1] &&
				data[0]!=0x00 &&
				data[0]!=0x05 &&
				data[0]!=0x0A
			) flag=0;
		}
		code=0;
		for(i=0;i<3;i++){
			code *= 16;
			code += (int)data[2-i];
		}
	}
	if(SerialEN){
		Serial.print("DecCode = 0x");
		for(i=0;i<3;i++) Serial.print(data[2-i],HEX);
		Serial.print(" -> "); Serial.println(code,DEC);
	}
	return(code);
}

byte ir2dec(byte *data){	// data長＝6バイト
// 赤外線リモコン信号を解析する(SHARP AQUOS / Panasonic VIERA)
	byte ret=0xFF;
	if(data[0]==0xAA && data[1]==0x5A && data[2]==0x8F && data[3]==0x12){	// Sharp
		ret=(byte)'\n';
		if(data[4] >= 0x01 && data[4] <= 0x09 && (data[5]&0xF) == 0x01) ret = data[4];
		if(data[4] >= 0x4E && data[4] <= 0x56 && (data[5]&0xF) == 0x02) ret = data[4] - 0x4D;
		if(data[4] == 0x0A && (data[5]&0xF) == 0x01) ret = 0x00;
		if(data[4] == 0x57 && (data[5]&0xF) == 0x02) ret = 0x00;
		if(data[4] == 0x11) ret=(byte)'+';
		if(data[4] == 0x12) ret=(byte)'-';
		if(data[4] == 0x80 || data[4] == 0x61) ret=(byte)'B';
		if(data[4] == 0x81 || data[4] == 0x62) ret=(byte)'R';
		if(data[4] == 0x82 || data[4] == 0x63) ret=(byte)'G';
		if(data[4] == 0x83 || data[4] == 0x64) ret=(byte)'Y';
	}
	if(data[0]==0x02 && data[1]==0x20 && data[2]==0x80 && data[3]==0x00){	// Panasonic
		ret=(byte)'\n';
		if(data[4] >= 0x10 && data[4] <= 0x18 && (data[5]&0xF0) == 0x90) ret = data[4] - 0x0F;
		if(data[4] == 0x19 && data[5] == 0x99) ret = 0x00;
		if(data[4] == 0x34) ret=(byte)'+';
		if(data[4] == 0x35) ret=(byte)'-';
		if(data[4] == 0x73) ret=(byte)'B';
		if(data[4] == 0x70) ret=(byte)'R';
		if(data[4] == 0x71) ret=(byte)'G';
		if(data[4] == 0x72) ret=(byte)'Y';
	}
	if(SerialEN){
		Serial.print("RemCode = ");
		for(byte i=0;i<6;i++){
			if(data[i]<0x10)Serial.print('0');
			Serial.print(data[i],HEX); Serial.print(' ');
		}
		Serial.print("-> ");
		if(ret<10) Serial.println(ret,DEC);
		else if(ret==(byte)'\n')Serial.println("\\n");
		else Serial.println((char)ret);
	}
	return(ret);
}

byte getSwCode(){
	byte in=0xFF,buf=0x00;
	while(in){
		in =  (byte)digitalRead(SW3_PIN)
			+ ((byte)(digitalRead(SW4_PIN))<<1)
			+ ((byte)(digitalRead(SW5_PIN))<<2)
			+ ((byte)(digitalRead(SW6_PIN))<<3);
		buf |= in;
	}
	return(buf);
}

boolean getSwDir(){
	boolean dir=0;
	unsigned long timer;
	byte i, data[4];
	int code;
	for(i=0;i<4;i++) data[i]=0x00;
	for(i=0;i<4;i++){
		timer = millis() + 1500;
		delay(10);
		while( millis() < timer && !data[i]) data[i] = 0x0F & getSwCode();
		if(!data[i])break;
	}
	code = (((int)data[0])<<8) + (((int)data[1])<<4) + (int)data[2];
	if(i>0) i--;
	if(data[i]==0x0A){		// 最後のコードが0x0Aの場合に挿入と判断する
		dir=1;
		if(i==3) yokCode = code;
	}
	if(SerialEN){
		Serial.print("InpCode = 0x");
		for(i=0;i<3;i++) Serial.print(data[i],HEX);
		Serial.print(" -> "); Serial.println(code,DEC);
		Serial.print("Medal   = ");
		Serial.println(dir,DEC);
		delay(50);
	}
	return(dir);
}

void error(){
	while(1){digitalWrite(LED_PIN,1);delay(200);digitalWrite(LED_PIN,0);delay(200);}
}

void setup(){
	pinMode(LED_PIN,OUTPUT); 
	pinMode(IR_VCC_PIN,OUTPUT);
	digitalWrite(IR_VCC_PIN,HIGH);
	yok_init();
	port_off();
	if(!(IrRemoEN|SerialEN)) error();
	if(IrRemoEN) ir_read_init();
	if(SerialEN){
		Serial.begin(9600);
		Serial.println("IR to Yok Code Converter by Wataru KUNINO");
		Serial.print("YokType = ");
		if(YokTypeWhite)Serial.println("Wite");
		if(YokTypeZero) Serial.println("Zero");
		delay(100);
	}
}

void loop(){
	byte data[IR_DATA_LEN];
	byte keyIn=0xFF;
	
	digitalWrite(LED_PIN,HIGH);
	if(IrRemoEN && !ardIsrSrc){
		if(ir_read(data,IR_DATA_LEN,0) > 8){		// 8ビット以上の赤外線データを読み込む
			keyIn = ir2dec(data);
		}
		do delay(100);								// 受信信号が無くなるまで待つ
		while(ir_read(data,1,0)>0);					// (チャタリング防止のため)
		ardIsrSrc=0xFF;
		time = 0;
	}
	if(SerialEN){
		if(Serial.available()) keyIn = (byte)Serial.read() - (byte)'0';
	}
	if(IrRemoEN && ardIsrSrc==2){					// WDT起動
		if(SerialEN){
			Serial.print("WDTimer = ");	Serial.print(time,DEC);
			Serial.print("/");	Serial.println(AutoPowerSave,DEC);
			delay(20);
		}
		if(time >= AutoPowerSave){					// オートパワーオフ
			digitalWrite(IR_VCC_PIN,LOW);
			digitalWrite(LED_PIN,LOW);
			if(SerialEN){
				Serial.println("Power Off");
				delay(100);
			}
			disable_watchdog();
			detachInterrupt(0);
			attachInterrupt(1, ardIsrSw, CHANGE);	// Digital 3を割込み入力に設定
			interrupts();
			system_sleep();
			digitalWrite(IR_VCC_PIN,HIGH);
			digitalWrite(LED_PIN,HIGH);
			time = 0;
		}
	}
	if(ardIsrSrc==1){								// メダル挿入の検出
		yokCodeMedal=getSwDir();					// メダル方向を取得
		time = 0;
	}
	
	if(keyIn != 0xFF){
		if(keyIn < 10){								// 数字入力時
			keyValue *= 10;
			keyValue += (int)keyIn;
			keyLength++;
		}else{
			keyLength = 0; keyValue = 0;
			if(yokCodeMedal){
				code_out(yokCode,0); delay(1500);		// メダル取り外し実行
				yokCodeMedal = 0;						// メダル空状態に設定
			}
			switch(keyIn){
				case (byte)'+':
					yokCode=code_increment(yokCode);	// 次のコードを取得
					code_out(yokCode,1);				// メダル挿入の実行
					yokCodeMedal = 1;					// メダル挿入状態に設定
					break;
				case (byte)'-':
					yokCode=code_decrement(yokCode);	// 前のコードを取得
					code_out(yokCode,1);				// メダル挿入の実行
					yokCodeMedal = 1;					// メダル挿入状態に設定
					break;
				case (byte)'B':
					yokCode=2382;						// 白：不用心
					if(YokTypeZero) yokCode=3665;		// 零：猫型ロボット
					code_out(yokCode,1);				// メダル挿入の実行
					yokCodeMedal = 1;					// メダル挿入状態に設定
					break;
				case (byte)'R':
					yokCode=2913;						// 白：自爆の猫
					if(YokTypeZero) yokCode=3090;		// 零：獅子
					code_out(yokCode,1);				// メダル挿入の実行
					yokCodeMedal = 1;					// メダル挿入状態に設定
					break;
				case (byte)'G':
					yokCode=3069;						// 白：どんより
					if(YokTypeZero) yokCode=3484;		// 零：自爆の猫
					code_out(yokCode,1);				// メダル挿入の実行
					yokCodeMedal = 1;					// メダル挿入状態に設定
					break;
				case (byte)'Y':
					yokCode=3470;						// 白：武士の猫
					if(YokTypeZero) yokCode=339;		// 零：竜型ロボット
					code_out(yokCode,1);				// メダル挿入の実行
					yokCodeMedal = 1;					// メダル挿入状態に設定
					break;
				case (byte)'\n':						// 実行コマンドの入力時
				default:
					break;
			}
		}
	}
	if(keyLength >= 4){								// 4桁入力
		if(yokCodeMedal){							// メダル挿入状態時
			code_out(yokCode,0); delay(1500);		// メダル取り外し実行
		}
		yokCode = keyValue;
		code_out(yokCode,1);						// メダル挿入の実行
		yokCodeMedal = 1;							// メダル挿入状態に設定
		keyLength = 0; keyValue = 0;
	}
	
	/* Arduino電源オフ */
	digitalWrite(LED_PIN,LOW);
	if(IrRemoEN){
		setup_watchdog(9);							// WDTを割り込みモード(8秒)で起動
		attachInterrupt(0, ardIsrIr, CHANGE);		// Digital 2を割込み入力に設定
		attachInterrupt(1, ardIsrSw, CHANGE);		// Digital 3を割込み入力に設定
		interrupts();
		system_sleep();								// Arduino パワーダウン
	}
}

void ardIsrIr(){									// 赤外線信号による割込み処理
	noInterrupts();
	ardIsrSrc=0;
}

void ardIsrSw(){									// 実メダル挿入による割込み処理
	noInterrupts();
	ardIsrSrc=1;
}

ISR(WDT_vect) {										// WDTがタイムアップした時に実行される処理
	noInterrupts();
	time++;
	ardIsrSrc=2;
}
